/*
 * Copyright (c) 2005, Swedish Institute of Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack
 *
 * @(#)$Id: dhcpc.c,v 1.2 2006/06/11 21:46:37 adam Exp $
 */

#include "includes.h"

#define STATE_INITIAL         0
#define STATE_SENDING         1
#define STATE_OFFER_RECEIVED  2
#define STATE_CONFIG_RECEIVED 3
#define STATE_FAIL            4

#define HOSTNAME "Netgear Camera model no."
#define VENDORID "Netgear"

#define BOOTP_BROADCAST 0x8000
#define BOOTP_UNICAST 0x0000

#define UIP_ARP_DISABLE_REPLY_ALL	0
#define UIP_ARP_DISABLE_REPLY_DHCPSRV	1
#define UIP_ARP_ENABLE_REPLY_ALL	2

static u8_t xid[4] = {0xad, 0xde, 0x12, 0x23};
static const u8_t magic_cookie[4] = {99, 130, 83, 99};

struct s_dhcp_state g_dhcp_state;

/*---------------------------------------------------------------------------*/
	static u8_t *
add_server_id(u8_t *optptr)
{
	*optptr++ = DHCP_OPTION_SERVER_ID;
	*optptr++ = 4;
	memcpy(optptr, g_dhcp_state.serverid, 4);
	return optptr + 4;
}
/*---------------------------------------------------------------------------*/
	static u8_t *
add_req_ipaddr(u8_t *optptr)
{
	*optptr++ = DHCP_OPTION_REQ_IPADDR;
	*optptr++ = 4;
	memcpy(optptr, g_dhcp_state.ipaddr, 4);
	return optptr + 4;
}
/*---------------------------------------------------------------------------*/
	static u8_t *
add_req_hostname(u8_t *optptr)
{
	*optptr++ = DHCP_OPTION_HOST_NAME;
	*optptr++ = strlen(HOSTNAME);
	memcpy(optptr, HOSTNAME, strlen(HOSTNAME));
	return optptr + strlen(HOSTNAME);	
}
/*---------------------------------------------------------------------------*/
	static u8_t *
add_req_vendor_id(u8_t *optptr)
{
	*optptr++ = DHCP_OPTION_VENDOR_ID;
	*optptr++ = strlen(VENDORID);
	memcpy(optptr, VENDORID, strlen(VENDORID));
	return optptr + strlen(VENDORID);
}
/*---------------------------------------------------------------------------*/
	static u8_t *
add_req_options(u8_t *optptr)
{
	*optptr++ = DHCP_OPTION_REQ_LIST;
	*optptr++ = 9;
	*optptr++ = DHCP_OPTION_SUBNET_MASK;
	*optptr++ = DHCP_OPTION_STATICROUTER;
	*optptr++ = DHCP_OPTION_ROUTER;
	*optptr++ = DHCP_OPTION_DNS_SERVER;
	*optptr++ = DHCP_OPTION_DOMIN_NAME;
	*optptr++ = DHCP_OPTION_BROADCAST;
	*optptr++ = DHCP_OPTION_LEASE_TIME;
	*optptr++ = DHCP_OPTION_RENEWAL;
	*optptr++ = DHCP_OPTION_REBINDING;

	return optptr;
}
/*---------------------------------------------------------------------------*/
	static void
//create_msg(register struct dhcp_msg *m)
create_msg(register struct dhcp_msg *m,u16 broadcast_flag)
{
	uip_ipaddr_t addr;
	uip_ipaddr(addr, 0,0,0,0);

	m->op = DHCP_REQUEST;
	m->htype = DHCP_HTYPE_ETHERNET;
	m->hlen = g_dhcp_state.mac_len;
	m->hops = 0;
	memcpy(m->xid, xid, sizeof(m->xid));
	m->secs = 0;
	//m->flags = HTONS(BOOTP_BROADCAST); /*  Broadcast bit. */
	m->flags = HTONS(broadcast_flag); /*  Broadcast bit. */
	/*  uip_ipaddr_copy(m->ciaddr, uip_hostaddr);*/
//	memcpy(m->ciaddr, uip_hostaddr[uip_dev_index], sizeof(m->ciaddr));
	memcpy(m->ciaddr, addr, sizeof(m->ciaddr));
	memset(m->yiaddr, 0, sizeof(m->yiaddr));
	memset(m->siaddr, 0, sizeof(m->siaddr));
	memset(m->giaddr, 0, sizeof(m->giaddr));
	memcpy(m->chaddr, g_dhcp_state.mac_addr, g_dhcp_state.mac_len);
	memset(&m->chaddr[g_dhcp_state.mac_len], 0, sizeof(m->chaddr) - g_dhcp_state.mac_len);
#ifndef UIP_CONF_DHCP_LIGHT
	memset(m->sname, 0, sizeof(m->sname));
	memset(m->file, 0, sizeof(m->file));
#endif

	memcpy(m->options, magic_cookie, sizeof(magic_cookie));
}
/*---------------------------------------------------------------------------*/
	static void
send_discover(void)
{
	u8_t *end;
	struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;

	//create_msg(m);
	create_msg(m,BOOTP_BROADCAST);

	end = dhcp_add_msg_type(&m->options[4], DHCPDISCOVER);
	end = add_req_options(end);
	end = add_req_hostname(end);
	end = dhcp_add_end(end);

	uip_send(uip_appdata, end - (u8_t *)uip_appdata);

}
/*---------------------------------------------------------------------------*/

	static void
send_request(u16 broadcastflag)
{
	u8_t *end;
	struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;

	//create_msg(m);
	create_msg(m,broadcastflag);
	end = dhcp_add_msg_type(&m->options[4], DHCPREQUEST);
	//if(!broadcastflag)
	if(broadcastflag==BOOTP_UNICAST)
		end = add_server_id(end);
	end = add_req_ipaddr(end);
	end = add_req_hostname(end);
	end = add_req_options(end);
	end = dhcp_add_end(end);

	uip_send(uip_appdata, end - (u8_t *)uip_appdata);
}
static u8_t parse_options(u8_t *optptr, int len)
{
	u8_t *end = optptr + len;
	u8_t type = 0;
	while(optptr < end) {
		switch(*optptr) {
			case DHCP_OPTION_SUBNET_MASK:
				memcpy(g_dhcp_state.netmask, optptr + 2, 4);
				break;
			case DHCP_OPTION_ROUTER:
				memcpy(g_dhcp_state.default_router, optptr + 2, 4);
				break;
			case DHCP_OPTION_DNS_SERVER:
				memcpy(g_dhcp_state.dnsaddr, optptr + 2, 4);
				break;
			case DHCP_OPTION_MSG_TYPE:
				type = *(optptr + 2);
				break;
			case DHCP_OPTION_SERVER_ID:
				memcpy(g_dhcp_state.serverid, optptr + 2, 4);
				break;
			case DHCP_OPTION_LEASE_TIME:
				memcpy(g_dhcp_state.lease_time, optptr + 2, 4);
				g_dhcp_state.leasetime = ntohs(g_dhcp_state.lease_time[0])*65536 + ntohs(g_dhcp_state.lease_time[1]);
				break;
			case DHCP_OPTION_END:
				return type;
		}

		optptr += optptr[1] + 2;
	}
	return type;
}
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
	static u8_t
parse_msg(void)
{
	struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;
	uip_markolddata();
	if(m->op == DHCP_REPLY &&
			memcmp(m->xid, xid, sizeof(xid)) == 0 &&
			memcmp(m->chaddr, g_dhcp_state.mac_addr, g_dhcp_state.mac_len) == 0) {
		memcpy(g_dhcp_state.ipaddr, m->yiaddr, 4);
		xid[2]+=random_out()%256;xid[3]++;
		return parse_options(&m->options[4], uip_datalen());
	}

	return 0;
}

/*---------------------------------------------------------------------------*/

static int dhcp_try_num = 0;
#define DS_FAIL_COUNT  15

static PT_THREAD(handle_dhcp(int pMaxDHCPConnectAttempts))
{
	u8 ret;
	uip_ipaddr_t addr;
	static int timeout_num;
	PT_BEGIN(&g_dhcp_state.pt);
	//uip_arp_reply_enable(1);
	uip_arp_reply_enable(UIP_ARP_DISABLE_REPLY_DHCPSRV);
	if(!(g_dhcp_state.ipaddr[0]==0&&g_dhcp_state.ipaddr[1]==0)){
		g_dhcp_state.dhcp_ticks = CLOCK_SECOND*3;
		//debug_printf("send request %d, %d\n",g_dhcp_state.dhcp_ticks,ticks);
		//send_request(1);
		send_request(BOOTP_UNICAST);
		timer_set(&g_dhcp_state.timer, g_dhcp_state.dhcp_ticks);
		do{
			PT_WAIT_UNTIL(&g_dhcp_state.pt, uip_newdata() || timer_expired(&g_dhcp_state.timer));
			if(uip_newdata())
			{
				ret = parse_msg();
				if(ret == DHCPACK) {
					debug_printf("Attempt to get an IP address with DHCP[1]\n");
					g_dhcp_state.state = STATE_CONFIG_RECEIVED;
					goto config;
				}else if(ret == DHCPNAK) {
					//debug_printf("get dhcpnack\n");
					break;
				}
				//debug_printf("get vaild data\n");
				continue;
			}
			//debug_printf("+dhcp request timer expired %d\n",ticks);
			break;
		}while(1);
	}
restart:
	//uip_arp_reply_enable(0);
	dhcp_try_num++;
	if ( dhcp_try_num > pMaxDHCPConnectAttempts)
	{
		debug_printf("Attempt to get an IP address with DHCP fail %d times,exit.\n",pMaxDHCPConnectAttempts);
		dhcp_try_num = 0;
		g_dhcp_state.state = STATE_FAIL;//˳־
		dhcpc_configured(&g_dhcp_state);
		return 0;
	}
	debug_printf("Attempt to get an IP address with DHCP[%d]\n",dhcp_try_num);

	uip_arp_reply_enable(UIP_ARP_DISABLE_REPLY_ALL);
	/* try_again:*/
	uip_ipaddr(addr, 255,255,255,255);
	uip_udp_set_dst(g_dhcp_state.conn,addr,HTONS(DHCPC_SERVER_PORT));
	g_dhcp_state.state = STATE_SENDING;
	g_dhcp_state.dhcp_ticks = CLOCK_SECOND;
	do {
		//debug_printf("send_discover %d\n",ticks);
		send_discover();
		timer_set(&g_dhcp_state.timer, g_dhcp_state.dhcp_ticks);
looper:
		PT_WAIT_UNTIL(&g_dhcp_state.pt, uip_newdata() || timer_expired(&g_dhcp_state.timer));

		if(uip_newdata())
		{
			if(parse_msg() == DHCPOFFER)
			{
				g_dhcp_state.state = STATE_OFFER_RECEIVED;
				break;
			}
			else
			{
				goto looper;
			}
		}else
		{
			debug_printf("-dhcpc discover timer expired\n");
		}

		if(g_dhcp_state.dhcp_ticks < CLOCK_SECOND * 10)
		{
			g_dhcp_state.dhcp_ticks += CLOCK_SECOND;
		}

		if ( dhcp_try_num == pMaxDHCPConnectAttempts)
		{
			goto restart;
		}
		dhcp_try_num++;
		debug_printf("Attempt to get an IP address with DHCP[%d]\n",dhcp_try_num);
	} while(g_dhcp_state.state != STATE_OFFER_RECEIVED);

request:
	if(g_dhcp_state.needack){
		g_dhcp_state.dhcp_ticks = CLOCK_SECOND;

		do {
			//debug_printf("send_request %d\n",ticks);
			//send_request(0);
			send_request(BOOTP_BROADCAST);
			timer_set(&g_dhcp_state.timer, g_dhcp_state.dhcp_ticks);
looper1:
			PT_WAIT_UNTIL(&g_dhcp_state.pt, uip_newdata() || timer_expired(&g_dhcp_state.timer));
			if(uip_newdata())
			{
				ret = parse_msg();
				if( ret == DHCPACK) {
					g_dhcp_state.state = STATE_CONFIG_RECEIVED;
					break;
				}else if(ret == DHCPNAK) {
					//debug_printf("get dhcpnack\n");
					goto restart;
				}
				//debug_printf("get vaild data\n");
				goto looper1;
			}

			debug_printf("+dhcp request timer expired %d\n",ticks);
			if ( dhcp_try_num == pMaxDHCPConnectAttempts )
			{
				goto restart;
			}
			dhcp_try_num++;
			debug_printf("Attempt to get an IP address with DHCP[%d]\n",dhcp_try_num);

			if(g_dhcp_state.dhcp_ticks <= CLOCK_SECOND * 10) {
				g_dhcp_state.dhcp_ticks += CLOCK_SECOND;
			} else {
				goto restart;
			}
		} while(g_dhcp_state.state != STATE_CONFIG_RECEIVED);
	}

config:
#if 1
//	debug_printf("Got IP address %d.%d.%d.%d\n",
//			uip_ipaddr1(g_dhcp_state.ipaddr), uip_ipaddr2(g_dhcp_state.ipaddr),
//			uip_ipaddr3(g_dhcp_state.ipaddr), uip_ipaddr4(g_dhcp_state.ipaddr));
//	debug_printf("Got netmask %d.%d.%d.%d\n",
//			uip_ipaddr1(g_dhcp_state.netmask), uip_ipaddr2(g_dhcp_state.netmask),
//			uip_ipaddr3(g_dhcp_state.netmask), uip_ipaddr4(g_dhcp_state.netmask));
//	debug_printf("Got DNS server %d.%d.%d.%d\n",
//			uip_ipaddr1(g_dhcp_state.dnsaddr), uip_ipaddr2(g_dhcp_state.dnsaddr),
//			uip_ipaddr3(g_dhcp_state.dnsaddr), uip_ipaddr4(g_dhcp_state.dnsaddr));
//	debug_printf("Got default router %d.%d.%d.%d\n",
//			uip_ipaddr1(g_dhcp_state.default_router), uip_ipaddr2(g_dhcp_state.default_router),
//			uip_ipaddr3(g_dhcp_state.default_router), uip_ipaddr4(g_dhcp_state.default_router));
	debug_printf("DHCP Lease expires in %d seconds\n",g_dhcp_state.leasetime);
#endif
	//uip_arp_reply_enable(2);
	uip_arp_reply_enable(UIP_ARP_ENABLE_REPLY_ALL);
	dhcpc_configured(&g_dhcp_state);

	if(1)
	{
		timeout_num = 0;
		timer_set(&g_dhcp_state.timer, 51*g_dhcp_state.leasetime);
		debug_printf("dhcp ok\n");
		do{

			PT_WAIT_UNTIL(&g_dhcp_state.pt, uip_newdata() || timer_expired(&g_dhcp_state.timer));
			if(uip_newdata())
			{
				ret = parse_msg();
				if(ret  == DHCPACK)
				{
					//debug_printf("get dhcp ack\n");
					//uip_arp_reply_enable(2);
					uip_arp_reply_enable(UIP_ARP_ENABLE_REPLY_ALL);
					timer_set(&g_dhcp_state.timer, 51*g_dhcp_state.leasetime);
					timeout_num = 0;
				}else if(ret == DHCPNAK)
				{
					//debug_printf("get dhcp nack\n");
					goto restart;
				}
				//debug_printf("get vaild data\n");
				continue;
			}else if(timer_expired(&g_dhcp_state.timer)){
				if(timeout_num==0)
				{
					debug_printf("send unicast dhcp request for lease time, %d\n",ticks);
					uip_udp_set_dst(g_dhcp_state.conn,g_dhcp_state.serverid,HTONS(DHCPC_SERVER_PORT));
					//uip_arp_reply_enable(1);
					//send_request(0);
					uip_arp_reply_enable(UIP_ARP_DISABLE_REPLY_DHCPSRV);
					send_request(BOOTP_UNICAST);					
					timer_set(&g_dhcp_state.timer, 37*g_dhcp_state.leasetime);
				}else if(timeout_num==1)
				{
					debug_printf("send broadcast dhcp request for lease time, %d\n",ticks);
					uip_ipaddr(addr, 255,255,255,255);
					uip_udp_set_dst(g_dhcp_state.conn,addr,HTONS(DHCPC_SERVER_PORT));
					//uip_arp_reply_enable(0);
					//send_request(1);
					uip_arp_reply_enable(UIP_ARP_DISABLE_REPLY_ALL);
					send_request(BOOTP_BROADCAST);					
					timer_set(&g_dhcp_state.timer, 12*g_dhcp_state.leasetime);
				}else
				{
					debug_printf("dhcp lease timeout, re-request dhcp, %d\n",ticks);
					PT_RESTART(&g_dhcp_state.pt);
				}
				timeout_num++;
			}

		}while(1);
	}

	if(g_dhcp_state.conn != NULL)
	{
		uip_udp_remove(g_dhcp_state.conn);
	}
	/*  timer_stop(&g_dhcp_state.timer);*/

	/*
	 * PT_END restarts the thread so we do this instead. Eventually we
	 * should reacquire expired leases here.
	 */
	//jacky marked no need hang up thread
	while(0) {
		PT_YIELD(&g_dhcp_state.pt);
	}
	PT_END(&g_dhcp_state.pt);
}

/*---------------------------------------------------------------------------*/
int dhcpc_init(const void *mac_addr, int mac_len, u8 needack, int dev_index, int (*handle)(int))
{
	uip_ipaddr_t addr;

	g_dhcp_state.mac_addr = mac_addr;
	g_dhcp_state.mac_len  = mac_len;
	g_dhcp_state.needack = needack;
	uip_gethostaddr(dev_index, g_dhcp_state.ipaddr);
	uip_getdsaddr(dev_index, g_dhcp_state.serverid);
	g_dhcp_state.state = STATE_INITIAL;
	xid[2]+=random_out()%256;xid[3]++;
	uip_ipaddr(addr, 255,255,255,255);
	g_dhcp_state.conn = uip_udp_new(&addr, HTONS(DHCPC_SERVER_PORT), dev_index, handle);
	if(g_dhcp_state.conn != NULL) {
		uip_udp_bind(g_dhcp_state.conn, HTONS(DHCPC_CLIENT_PORT));
		debug_printf("get new udp conn %x\n",g_dhcp_state.conn);
	}
	else
	{
		debug_printf("dhcpc_init error\n");
		//		while(1);
		return 2;
	}
	PT_INIT(&g_dhcp_state.pt);
	return 0;
}
void dhcpc_close()
{
	if(g_dhcp_state.conn != NULL)
	{
		uip_udp_remove(g_dhcp_state.conn);
	}
	return ;
}
/*---------------------------------------------------------------------------*/
void dhcpc_appcall(int pMaxDHCPConnectAttempts)
{
	handle_dhcp(pMaxDHCPConnectAttempts);
}

/*---------------------------------------------------------------------------*/
void dhcpc_request(void)
{
	u16_t ipaddr[2];

	if(g_dhcp_state.state == STATE_INITIAL) {
		//uip_ipaddr(ipaddr, 0,0,0,0);
		//uip_sethostaddr(g_dhcp_state.conn->dev_index,ipaddr);
		//handle_dhcp();
	}
}
/*---------------------------------------------------------------------------*/
