#include "includes.h"

#define MAX_REQUEST_LIST 16
static unsigned char request_list[MAX_REQUEST_LIST];
static int request_list_cnt;

#define MAX_CLIENT_LIST 16 //must <= 32 , because we ust bitmap.
//also must be 2^, because we use '%' to hash to cur_dev_list.
#define CLIENT_IP_START 33 //start from xxx.xxx.xxx.33
static struct
{
	u8_t macaddr[8]; //0~5: six bytes mac. 6: 1-used,0-notused.
}cur_dev_list[MAX_CLIENT_LIST];

static u8_t get_ip_from_mac(u8 * macaddr)
{
	u8_t i;

	for(i=0; i < MAX_CLIENT_LIST; i++){
		if(cur_dev_list[i].macaddr[6] == 0){
			break;
		}
		if(! memcmp(&cur_dev_list[i].macaddr[0], macaddr, 6)){
			break;
		}
	}
	if(i >= MAX_CLIENT_LIST){
		i = macaddr[5] % MAX_CLIENT_LIST ;
	}

	memcpy(&cur_dev_list[i].macaddr[0], macaddr, 6);
	cur_dev_list[i].macaddr[6] = 1;

	debug_printf("DHCPServer offer IP: 192.168.2.%d\n", CLIENT_IP_START + i );
	return CLIENT_IP_START + i;
}

/*
#define MAX_VENDOR_INFO 32
static unsigned char vendor_info[MAX_VENDOR_INFO];
static int vendor_info_length;
*/
static u8_t parse_options(u8_t *optptr, int len)
{
  u8_t *end = optptr + len;
  u8_t type = 0;

  request_list_cnt = 0;
//  vendor_info_length;

  while(optptr < end) {
    switch(*optptr) {
			/*
	case DHCP_OPTION_VENDOR_ID:
      vendor_info_length = optptr[1];
	  if(vendor_info_length > MAX_VENDOR_INFO){
		vendor_info_length = MAX_VENDOR_INFO;
	  }
	  memcpy(vendor_info, optptr + 2, vendor_info_length);
	  break;
	  */
	case DHCP_OPTION_REQ_LIST:
      request_list_cnt = optptr[1];
	  if(request_list_cnt > MAX_REQUEST_LIST){
		request_list_cnt = MAX_REQUEST_LIST;
	  }
	  memcpy(request_list, optptr + 2, request_list_cnt);
      break;
    case DHCP_OPTION_MSG_TYPE:
      type = *(optptr + 2);
      break;
    case DHCP_OPTION_END:
      return type;
    }
    optptr += optptr[1] + 2;
  }
  return type;
}

int dhcpserver_handle(int event)
{
	if(uip_udp_conn->lport != 67){
		return 0;
	}

	if(!uip_newdata()){
		return -1;
	}

  struct dhcp_msg *m = (struct dhcp_msg *)uip_appdata;
  u8_t *end;

  u8_t type = parse_options(&m->options[4], uip_datalen());
  if(type == DHCPDISCOVER || type == DHCPREQUEST){
    m->yiaddr[0] = 192;
    m->yiaddr[1] = 168;
    m->yiaddr[2] = 2;
    m->yiaddr[3] = get_ip_from_mac(&m->chaddr[0]) ;
    m->siaddr[0] = 192;
    m->siaddr[1] = 168;
    m->siaddr[2] = 2;
    m->siaddr[3] = 103;
    m->op = DHCP_REPLY;

    if(type == DHCPDISCOVER){
      end = dhcp_add_msg_type(&m->options[4], DHCPOFFER);
    }else{
      end = dhcp_add_msg_type(&m->options[4], DHCPACK);
    }

	int i;
	for(i=0; i<request_list_cnt; i++){
		switch(request_list[i]){
			case DHCP_OPTION_SUBNET_MASK:
				*end++ = request_list[i];
				*end++ = 4;
				*end++ = 255;
				*end++ = 255;
				*end++ = 255;
				*end++ = 0;
				break;
			case DHCP_OPTION_ROUTER:
			case DHCP_OPTION_DNS_SERVER:
				*end++ = request_list[i];
				*end++ = 4;
				*end++ = 192;
				*end++ = 168;
				*end++ = 2;
				*end++ = 103;
				break;
			case DHCP_OPTION_BROADCAST:
				*end++ = request_list[i];
				*end++ = 4;
				*end++ = 192;
				*end++ = 168;
				*end++ = 255;
				*end++ = 255;
				break;
#if 0
			case DHCP_OPTION_RENEWAL:
			case DHCP_OPTION_REBINDING:
			case DHCP_OPTION_LEASE_TIME:
				*end++ = request_list[i];
				*end++ = 4;
				*end++ = 0;
				*end++ = 1;
				*end++ = 0x51;
				*end++ = 0x80;
				break;
#endif
			default:
				debug_printf("DHCPserver don't support option:%d\n", request_list[i]);
				break;
		}
	}
	*end++ = DHCP_OPTION_SERVER_ID;
	*end++ = 4;
	*end++ = 192;
	*end++ = 168;
	*end++ = 2;
	*end++ = 103;

	*end++ = DHCP_OPTION_RENEWAL;
	*end++ = 4;
	*end++ = 0;
	*end++ = 1;
	*end++ = 0x51;
	*end++ = 0x80;

	*end++ = DHCP_OPTION_REBINDING;
	*end++ = 4;
	*end++ = 0;
	*end++ = 1;
	*end++ = 0x51;
	*end++ = 0x80;

	*end++ = DHCP_OPTION_LEASE_TIME;
	*end++ = 4;
	*end++ = 0;
	*end++ = 1;
	*end++ = 0x51;
	*end++ = 0x80;
	/*
	if(vendor_info_length){
		*end++ = DHCP_OPTION_VENDOR_INFO;
		*end++ = vendor_info_length;
		for(i=0; i<vendor_info_length; i++){
			*end++ = vendor_info[i];
		}
	}
	*/
    end = dhcp_add_end(end);

	uip_udp_conn->rport = 68;
extern u16_t uip_slen;
    uip_slen = end - (u8_t *)uip_appdata ;
  }

  return 1;

}

void dhcpserver_init(int dev_index)
{
	uip_ipaddr_t ipaddr;
	uip_ipaddr(ipaddr, 255,255,255,255);

	memset(cur_dev_list, 0, sizeof(cur_dev_list));

	/* listen on broadcast address, port 67 */
	uip_udp_listen(&ipaddr, 67, dev_index,dhcpserver_handle );
}
