#ifndef _OPENSSL_RTOS_H_
#define _OPENSSL_RTOS_H_

#include "includes.h"

//from opensslconf.h
#define OPENSSL_NO_EC_NISTP_64_GCC_128
#define OPENSSL_NO_GMP
#define OPENSSL_NO_JPAKE
#define OPENSSL_NO_KRB5
#define OPENSSL_NO_MD2
#define OPENSSL_NO_RC5
#define OPENSSL_NO_RFC3779
#define OPENSSL_NO_SCTP
#define OPENSSL_NO_STORE
#define OPENSSL_NO_DYNAMIC_ENGINE

#ifdef OPENSSL_ALGORITHM_DEFINES
#  define NO_EC_NISTP_64_GCC_128
#  define NO_GMP
#  define NO_JPAKE
#  define NO_KRB5
#  define NO_MD2
#  define NO_RC5
#  define NO_RFC3779
#  define NO_SCTP
#  define NO_STORE
#endif

#undef I386_ONLY
#undef OPENSSL_UNISTD
#undef OPENSSL_EXPORT_VAR_AS_FUNCTION

# define IDEA_INT unsigned int

# define MD2_INT unsigned int

# define RC2_INT unsigned int

# if !defined(RC4_INT)
/* using int types make the structure larger but make the code faster
 * on most boxes I have tested - up to %20 faster. */
/*
 * I don't know what does "most" mean, but declaring "int" is a must on:
 * - Intel P6 because partial register stalls are very expensive;
 * - elder Alpha because it lacks byte load/store instructions;
 */
#  define RC4_INT unsigned int
# endif
# if !defined(RC4_CHUNK)
/*
 * This enables code handling data aligned at natural CPU word
 * boundary. See crypto/rc4/rc4_enc.c for further details.
 */
#  undef RC4_CHUNK
# endif

# ifndef DES_LONG
#  define DES_LONG unsigned long
# endif

# undef BN_LLONG
# undef SIXTY_FOUR_BIT_LONG
# undef SIXTY_FOUR_BIT
# define THIRTY_TWO_BIT

#if defined(HEADER_RC4_LOCL_H) && !defined(CONFIG_HEADER_RC4_LOCL_H)
# define CONFIG_HEADER_RC4_LOCL_H
# undef RC4_INDEX
#endif

#if defined(HEADER_BF_LOCL_H) && !defined(CONFIG_HEADER_BF_LOCL_H)
# define CONFIG_HEADER_BF_LOCL_H
# undef BF_PTR
#endif /* HEADER_BF_LOCL_H */

#if defined(HEADER_DES_LOCL_H) && !defined(CONFIG_HEADER_DES_LOCL_H)
# define CONFIG_HEADER_DES_LOCL_H
# ifndef DES_DEFAULT_OPTIONS
/* the following is tweaked from a config script, that is why it is a
 * protected undef/define */
#ifndef DES_PTR
#undef DES_PTR
#endif

/* This helps C compiler generate the correct code for multiple functional
 * units.  It reduces register dependancies at the expense of 2 more
 * registers */
#ifndef DES_RISC1
#undef DES_RISC1
#endif

#ifndef DES_RISC2
#undef DES_RISC2
#endif

#if defined(DES_RISC1) && defined(DES_RISC2)
YOU SHOULD NOT HAVE BOTH DES_RISC1 AND DES_RISC2 DEFINED!!!!!
#endif

/* Unroll the inner loop, this sometimes helps, sometimes hinders.
 * Very mucy CPU dependant */
#ifndef DES_UNROLL
#undef DES_UNROLL
#endif
#endif //DES_DEFAULT_OPTIONS
#endif

//e_os2.h
#define OPENSSL_SYS_UNIX
#define OPENSSL_SYS_BA2

# define OPENSSL_UNISTD_IO OPENSSL_UNISTD
# define OPENSSL_DECLARE_EXIT /* declared in unistd.h */

# define OPENSSL_EXPORT extern
# define OPENSSL_IMPORT extern
# define OPENSSL_GLOBAL
#define OPENSSL_EXTERN OPENSSL_IMPORT

# define OPENSSL_IMPLEMENT_GLOBAL(type,name,value) OPENSSL_GLOBAL type _shadow_##name=value;
# define OPENSSL_DECLARE_GLOBAL(type,name) OPENSSL_EXPORT type _shadow_##name
# define OPENSSL_GLOBAL_REF(name) _shadow_##name

#  define ssize_t int

#ifndef ossl_ssize_t
#  define ossl_ssize_t ssize_t
#endif

//e_os.h
#  define MS_STATIC

#define get_last_socket_error()	errno
#define clear_socket_error()	errno=0
#define ioctlsocket(a,b,c)	ioctl(a,b,c)
#define closesocket(s)		close(s)
#define readsocket(s,b,n)	read((s),(b),(n))
#define writesocket(s,b,n)	write((s),(b),(n))

#  define MS_CALLBACK
#  define MS_FAR

#  define OPENSSL_NO_FP_API

#ifdef USE_SOCKETS
#      define OPENSSL_USE_IPV6 0
#endif

#define LIST_SEPARATOR_CHAR ':'

#    define OPENSSL_CONF    "openssl.cnf"
#    define SSLEAY_CONF     OPENSSL_CONF
#  define OPENSSL_EXIT(n) do { EXIT(n); return(n); } while(0)

/***********************************************/

#define DG_GCC_BUG	/* gcc < 2.6.3 on DGUX */

//my defines

#define OPENSSL_NO_STDIO
#define GETPID_IS_MEANINGLESS
#define NO_SYS_TYPES_H
#define OPENSSL_showfatal debug_printf
#define stderr NULL

#define OPENSSL_NO_ENGINE

#define time(X) ticks

int OPENSSL_strcasecmp(const char *str1, const char *str2);
int OPENSSL_strncasecmp(const char *str1, const char *str2, size_t n);
int OPENSSL_memcmp(const void *p1,const void *p2,size_t n);
# define OPENSSL_IMPLEMENTS_strncasecmp
# define strncasecmp OPENSSL_strncasecmp

# define OPENSSL_NO_POSIX_IO
struct sockaddr { int reserved; };

extern int errno;

#ifndef BUFSIZ
#define BUFSIZ 1024
#endif

#undef schedule
#undef vals
#undef check

typedef int FILE ;

#define strerror(num) "UNDEF ERR"

#define OPENSSLDIR ""

#define INT_MAX		0x7FFFFFFF

#define isspace(c) ((char)(c) == ' ')

#define isupper(x) (x >= 'A' && x <= 'Z')
#define tolower(x) (x - 'A' + 'a')

//represent stdlib.h
unsigned long int strtoul(const char *nptr, char **endptr, int base);
#define atoi util_atoi
void qsort(void *base, size_t nmemb, size_t size, int(*compar)(const void *, const void *));
#define exit(x) do{debug_printf("EXIT\n");}while(0)
#define getenv(x) NULL

s8 rtos_mem_init(void);
void *rtos_malloc(size_t size);
void rtos_free(void *memblock);
void *rtos_realloc(void *memblock, size_t size);


#endif
