#include "includes.h"

typedef struct os_mem {
    void   *OSMemAddr;
    void   *OSMemFreeList;
    u32  OSMemBlkSize;
    u32  OSMemNBlks;
    u32  OSMemNFree;
} OS_MEM;

static OS_MEM  *OSMemCreate (OS_MEM *pmem, void   *addr, u32  nblks, u32  blksize) {
    u8     *pblk; void     **plink; u32     i; plink = (void **)addr;
    pblk  = (u8 *)((u32)addr + blksize);
    for (i = 0; i < (nblks - 1); i++) { *plink = (void *)pblk; plink = (void **)pblk;
        pblk  = (u8 *)((u32)pblk + blksize); }
    *plink              = (void *)0;
    pmem->OSMemAddr     = addr; pmem->OSMemFreeList = addr; pmem->OSMemNFree    = nblks;
    pmem->OSMemNBlks    = nblks; pmem->OSMemBlkSize  = blksize; return (pmem);
} static void  *OSMemGet (OS_MEM  *pmem)
{ void      *pblk; if (pmem->OSMemNFree > 0) {
        pblk                = pmem->OSMemFreeList;
        pmem->OSMemFreeList = *(void **)pblk; pmem->OSMemNFree--;
        return (pblk); } return ((void *)0);
} static u8  OSMemPut (OS_MEM  *pmem, void    *pblk)
{ if (pmem->OSMemNFree >= pmem->OSMemNBlks) { return 1; }
    *(void **)pblk      = pmem->OSMemFreeList;
    pmem->OSMemFreeList = pblk; pmem->OSMemNFree++; return 0; }
static OS_MEM * rawos_pmem = NULL;
static unsigned int MAX_RAWOS_MEMP = 0;
static int rawos_mem_init(OS_MEM *pmem, unsigned int count) {
	rawos_pmem = pmem; MAX_RAWOS_MEMP = count; return 0; }
void *rtos_malloc(size_t size)
{ int i; void *pblk = NULL; 
    for (i = 0; i < MAX_RAWOS_MEMP; i++) {
        if ((rawos_pmem[i].OSMemBlkSize >= size) && (rawos_pmem[i].OSMemNFree > 0)) {
            break; } } 
    if (i == MAX_RAWOS_MEMP) { debug_printf("malloc failed\n"); return (void*)NULL; } 
    pblk = OSMemGet(&rawos_pmem[i]); if(pblk == NULL){ debug_printf("malloc failed\n"); } return pblk; } 
void rtos_free(void *memblock)
{ int i; for (i = 0; i < MAX_RAWOS_MEMP; i++) {
        if (((unsigned int)rawos_pmem[i].OSMemAddr + rawos_pmem[i].OSMemNBlks *rawos_pmem[i].OSMemBlkSize) > ((unsigned int)
            memblock) && ((unsigned int)rawos_pmem[i].OSMemAddr) < ((unsigned int)memblock + 1)) { break;
        } } if (i == MAX_RAWOS_MEMP)
    { return ; }
OSMemPut(&rawos_pmem[i], memblock); return ; }
void *rtos_calloc(size_t nmemb, size_t size)
{ void *ptr; ptr = rtos_malloc(nmemb *size);
    if (ptr) { memset(ptr, 0, nmemb *size); } return ptr;
} void *rtos_realloc(void *memblock, size_t size)
{ unsigned int i, old_blk_size; void *newblock; 
    for (i = 0; i < MAX_RAWOS_MEMP; i++) {
        if (((unsigned int)rawos_pmem[i].OSMemAddr + rawos_pmem[i].OSMemNBlks *rawos_pmem[i].OSMemBlkSize) > ((unsigned int)
            memblock) && ((unsigned int)rawos_pmem[i].OSMemAddr) < ((unsigned int)memblock + 1)) { break;
        } } if (i == MAX_RAWOS_MEMP) { return (void*)NULL; } 
    old_blk_size = rawos_pmem[i].OSMemBlkSize; if (old_blk_size >= size) {
        return memblock; } newblock = rtos_malloc(size);
    if (newblock == (void*)NULL) { return NULL; } 
    memcpy(newblock, memblock, old_blk_size); rtos_free(memblock); return newblock; }



#define MAX_OSMEM_PART 10
OS_MEM osmem_table[MAX_OSMEM_PART];

#define PRE_ALLOC( size, cnt ) \
			static u8 mem_pool_##size [ size * cnt ] ; \
            pmem = OSMemCreate(&osmem_table[osmem_table_index], mem_pool_##size , cnt, size); \
            if( pmem == NULL ){ \
                debug_printf("%d byte memory init failed\n", size); \
                return -1; \
            }\
			osmem_table_index ++; \
            if( osmem_table_index > MAX_OSMEM_PART) { \
                debug_printf("%d alloc: no osmem_table space\n", size); \
            	return -1; \
			}

s8 rtos_mem_init(void)
{
	memset(&osmem_table[0], 0, sizeof(osmem_table));

    OS_MEM *pmem = NULL;
	int osmem_table_index = 0;

    PRE_ALLOC(4, 0x20);
    PRE_ALLOC(8, 0x20);
    PRE_ALLOC(16, 0x20);
    PRE_ALLOC(32, 0x20);
    PRE_ALLOC(64, 0x20);
    PRE_ALLOC(128, 0x20);
    PRE_ALLOC(256, 0x20);
    PRE_ALLOC(512, 0x20);
    PRE_ALLOC(1024, 0x20);
    PRE_ALLOC(16384, 0x10);

	rawos_mem_init(&osmem_table[0], MAX_OSMEM_PART);

    return 0;
}
